# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

from i2fhirb2.i2b2model.metadata.commondimension import CommonDimension
from i2fhirb2.sqlsupport.dynobject import DynObject, DynElements


class ModifierDimension(CommonDimension):
    _t = DynElements(CommonDimension)

    @DynObject.entry(_t)
    def modifier_path(self) -> str:
        return self.path()

    @DynObject.entry(_t)
    def modifier_cd(self) -> str:
        return self.cd()

    @DynObject.entry(_t)
    def name_char(self) -> str:
        return self.name_char_()

    @DynObject.entry(_t)
    def modifier_blob(self) -> str:
        return self.blob()

    def __lt__(self, other):
        return self.modifier_path < other.modifier_path

    def __eq__(self, other):
        return self.modifier_path == other.modifier_path


class ModifierDimensionRoot(ModifierDimension):
    _t = DynElements(ModifierDimension)

    def __init__(self, base: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self._base = base

    @DynObject.entry(_t)
    def modifier_path(self) -> str:
        return '\\' + self._base + '\\'

    @DynObject.entry(_t)
    def modifier_cd(self) -> str:
        return self._base

    @DynObject.entry(_t)
    def name_char(self) -> str:
        return self._base + " root"

    @DynObject.entry(_t)
    def modifier_blob(self) -> str:
        return ''
